#ifndef PGC_H
#define PGC_H

#define PGC_MAX_EXPR_LENGTH 4096

/* The hashing functions that make the table lookups work may go slightly
   quicker if these values are powers of two. */

#define PGC_MAX_BINDINGS 256
#define PGC_CONF_TABLE_SIZE 256
#define PGC_FUNCTION_TABLE_SIZE 256
#define PGC_RECORD_TABLE_SIZE 256

enum pgc_types {
  PGC_NIL,
  PGC_INT, 
  PGC_FLOAT, 
  PGC_STRING,
  PGC_VECTOR,
  PGC_RECORD
};

enum pgc_error_types {
  PGC_ERROR_LEXICAL,
  PGC_ERROR_UNEXPECTED_EOF,
  PGC_ERROR_SYNTAX,
  PGC_ERROR_TYPE,
  PGC_ERROR_TYPE_NOT_COMPARABLE,
  PGC_ERROR_DIVIDE_BY_ZERO,
  PGC_ERROR_MODULO_ZERO,
  PGC_ERROR_UNBOUND_IDENTIFIER,
  PGC_ERROR_NAN,
  PGC_ERROR_VECTOR_TYPE,
  PGC_ERROR_SUBSCRIPT_RANGE,
  PGC_ERROR_SUBSCRIPT_TYPE,
  PGC_ERROR_RECORD_ASSIGNMENT,
  PGC_ERROR_RECORD_UNDEFINED_FIELD,
  PGC_ERROR_FUNCTION_ARG_TYPE,
  PGC_ERROR_FUNCTION_ARG_NUM,
  PGC_ERROR_FUNCTION_UNDEFINED,
  PGC_ERROR_RESULT_ASSIGNMENT
};

struct pgc_vector {
  int vector_type;
  int vector_length;
  void *vector;  
};

struct pgc_record {
  char *record_type_name; /* Use this to look up the salient details in the
			     record_table. */
  void **public_fields;
  void *struct_data;
  int reference_count;
};

extern int yy_flex_debug;

#endif
