#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_binding.h"
#include "pgc_hash.h"
#include "pgc_record.h"


void restore_result (struct pgc_environment *restore_env) {
  pgc_free_data (restore_env->result_type, restore_env->result);
  restore_env->result_type = 
    restore_env->top_level_result_type;
  restore_env->result = 
    pgc_copy_data (restore_env->top_level_result_type, 
		   restore_env->top_level_result);
  return;
}

void pgc_backup_result (struct pgc_environment *backup_env) {
  pgc_free_data (backup_env->top_level_result_type, 
		 backup_env->top_level_result);
  backup_env->top_level_result_type = 
    backup_env->result_type;
  backup_env->top_level_result = 
    pgc_copy_data (backup_env->result_type, backup_env->result);
  return;
}

struct pgc_environment *pgc_copy_environment 
(struct pgc_environment *old_environment) {
  int i;
  struct pgc_environment *new_environment = 
    (struct pgc_environment *) malloc (sizeof (struct pgc_environment));
  if (old_environment == NULL) {
    new_environment->result_type = PGC_NIL;
    new_environment->result = NULL;
    new_environment->top_level_result_type = PGC_NIL;
    new_environment->top_level_result = NULL;
    new_environment->bindings = 
      malloc (sizeof (struct pgc_binding *) * (PGC_MAX_BINDINGS));
    for (i = 0; i < PGC_MAX_BINDINGS; i++)
      new_environment->bindings[i] = NULL;
    new_environment->result_binding = NULL;
    return new_environment;
  }
  else {
    new_environment->result_type = old_environment->result_type;
    new_environment->result = 
      pgc_copy_data (old_environment->result_type, old_environment->result);
    new_environment->top_level_result_type = 
      old_environment->top_level_result_type;
    new_environment->top_level_result =
      pgc_copy_data (old_environment->top_level_result_type, 
		     old_environment->top_level_result);
    new_environment->bindings = 
      malloc (sizeof (struct pgc_binding *) * (PGC_MAX_BINDINGS));
    for (i = 0; i < PGC_MAX_BINDINGS; i++) {
      new_environment->bindings[i] = old_environment->bindings[i];
      if (old_environment->bindings[i] != NULL)
	old_environment->bindings[i]->reference_count++;
    }
    new_environment->result_binding = old_environment->result_binding;
    if (old_environment->result_binding != NULL)
      old_environment->result_binding->reference_count++;
    return new_environment;
  }
}

struct pgc_vector *pgc_copy_vector (struct pgc_vector *old_vector) {
  int i;
  struct pgc_vector *new_vector = malloc (sizeof (struct pgc_vector));
  new_vector->vector_type = old_vector->vector_type;
  new_vector->vector_length = old_vector->vector_length;
  new_vector->vector = NULL;
  if (new_vector->vector_length > 0) { 
    new_vector->vector = 
      malloc (sizeof (void *) * new_vector->vector_length);
    for (i = 0; i < new_vector->vector_length; i++)
      ((void **) new_vector->vector)[i] = 
	pgc_copy_data (new_vector->vector_type, 
		       ((void **) old_vector->vector)[i]);
  }
  return new_vector;
}

void *pgc_copy_data (int data_type, void *data) {
  void *new_data = NULL;
  switch (data_type) {
  case PGC_INT:
    (int *) new_data = (int *) malloc (sizeof (int));
    *((int *) new_data) = *((int *) data);
    break;
  case PGC_FLOAT:
    (double *) new_data = (double *) malloc (sizeof (double));
    *((double *) new_data) = *((double *) data);
    break;
  case PGC_STRING:
    (char *) new_data = strdup ((char *) data);
    break;
  case PGC_VECTOR:
    (struct pgc_vector *) new_data = 
      pgc_copy_vector ((struct pgc_vector *) data);
    break;
  case PGC_RECORD:
    (struct pgc_record *) new_data = (struct pgc_record *) data;
    ((struct pgc_record *) new_data)->reference_count++;
    break;
  default:
    break;			   
  }
  return new_data;
}

void pgc_free_vector (struct pgc_vector *dead_vector) {
  int i;
  struct pgc_vector **dead_subvector;
  if (dead_vector->vector_type == PGC_VECTOR) {
    dead_subvector = (struct pgc_vector **) dead_vector->vector;    
    for (i = 0; i < dead_vector->vector_length; i++)
      pgc_free_vector (dead_subvector[i]);
  }
  else
    free (dead_vector);
}

void pgc_delete_binding (struct pgc_binding *dead_binding) {
  dead_binding->reference_count--;
  if (dead_binding->reference_count < 1) {
    free (dead_binding->binding_name);
    pgc_free_data (dead_binding->value_type, dead_binding->value);
    free (dead_binding);
  }
  return;
}
  
void pgc_free_record (struct pgc_record *dead_record) {
  ((struct pgc_record *) dead_record)->reference_count--;
  if (((struct pgc_record *) dead_record)->reference_count == 0) {
    struct pgc_record_definition *scratch =
      pgc_lookup_record (dead_record->record_type_name);
    int i;
    if ((scratch != NULL) && (dead_record->struct_data != NULL)) {
      for (i = 0; i < scratch->field_count; i++)
	free (dead_record->public_fields[i]);
      free (dead_record->struct_data);
    }
  }
  return;
}

void pgc_delete_environment (struct pgc_environment *current_environment) {
  int i;
  pgc_free_data 
    (current_environment->result_type, current_environment->result);
  for (i = 0; i < PGC_MAX_BINDINGS; i++)
    if (current_environment->bindings[i] != NULL)
      pgc_delete_binding (current_environment->bindings[i]);
  free (current_environment->bindings);
  if (current_environment->result_binding != NULL)
    pgc_delete_binding (current_environment->result_binding);
  free (current_environment);
  return;
}

void pgc_free_data (int type, void *data) {
  if ((data != NULL) && (type != PGC_NIL)) {
    switch (type) {
    case PGC_VECTOR:
      pgc_free_vector (data);
      break;
    case PGC_RECORD:
      pgc_free_record (data);
      break;
    default:
      free (data);
      break;
    }
  }
  return;
}


/* Here're the definitions of the functions prototyped in our include... */
struct pgc_binding *pgc_lookup_binding 
(char *binding_name, struct pgc_environment *current_environment) {
  if (strcmp (binding_name, "result") == 0) {
    restore_result (current_environment);
    if (current_environment->result_binding != NULL)
      pgc_delete_binding (current_environment->result_binding);
    current_environment->result_binding = malloc (sizeof (struct pgc_binding));
    current_environment->result_binding->binding_name = strdup ("result");
    current_environment->result_binding->value_type =
      current_environment->result_type;
    current_environment->result_binding->value =
      pgc_copy_data (current_environment->result_type, 
		     current_environment->result);
    current_environment->result_binding->reference_count = 1;
    return current_environment->result_binding;
  }
  else {
    int i;
    int hashed_index = pgc_hash (binding_name, PGC_MAX_BINDINGS);
    if (current_environment->bindings[hashed_index] != NULL) {
      for (i = hashed_index; i < PGC_MAX_BINDINGS; i++) {
	if (current_environment->bindings[i] == NULL) {
	  return NULL;
	}
	if (strcmp (binding_name, 
		    current_environment->bindings[i]->binding_name) == 0)
	  return current_environment->bindings[i];
      }
      for (i = 0; i < hashed_index; i++) {
	if (current_environment->bindings[i] == NULL)
	  return NULL;
	if (strcmp (binding_name, 
		    current_environment->bindings[i]->binding_name) == 0)
	  return current_environment->bindings[i];
      }
      return NULL;
    }
    else
      return NULL;
  }
}

/* This function actually returns a new, malloc'd environment... */
struct pgc_environment *pgc_add_binding
(struct pgc_binding *new_binding, 
 struct pgc_environment *current_environment) {
  int i;
  int insertion_index = -1;
  int hashed_index = pgc_hash (new_binding->binding_name, PGC_MAX_BINDINGS);
  struct pgc_environment *new_environment = 
    pgc_copy_environment (current_environment);
  if (new_environment->bindings[hashed_index] != NULL) {
    for (i = hashed_index; i < PGC_MAX_BINDINGS; i++) {
      if ((new_environment->bindings[i] == NULL) ||
	  (strcmp (new_environment->bindings[i]->binding_name,
		   new_binding->binding_name) == 0)) {
	insertion_index = i;
	break;
      }
    }
    if (insertion_index == -1)
      for (i = 0; i < hashed_index; i++)
	if ((new_environment->bindings[i] == NULL) ||
	    (strcmp (new_environment->bindings[i]->binding_name,
		     new_binding->binding_name) == 0)) {
	  insertion_index = i;
	  break;
	}
  }
  if ((new_environment->bindings[hashed_index] == NULL) ||
      (insertion_index != -1)) {
    if (insertion_index == -1)
      insertion_index = hashed_index;
    /* There was a space for our object! */
    new_binding->reference_count++;
    new_environment->bindings[insertion_index] = new_binding;
    return new_environment;
  }
  else {
    printf ("unable to add binding!\n");
  }
  return current_environment;
}

