#include <math.h>
#include <string.h>
#include "pgc_hash.h"

/* Hash function from: [JENK97] Bob Jenkins. "Algorithm Alley: Hash Functions".
   Dr. Dobb's Journal. Semptember 1997. */
/* Okay, these hash routines require that we're working with 32-bit ints.  
   How to guarantee this? */

#define hashsize(n) ((int) 1 << (n))
#define hashmask(n) (hashsize (n) - 1)

#define PGC_HASH_MIX(a, b, c) { \
  a -= b; a -= c; a ^= (c >> 13); \
  b -= c; b -= a; b ^= (a << 8); \
  c -= a; c -= b; c ^= (b >> 13); \
  a -= b; a -= c; a ^= (c >> 12);  \
  b -= c; b -= a; b ^= (a << 16); \
  c -= a; c -= b; c ^= (b >> 5); \
  a -= b; a -= c; a ^= (c >> 3);  \
  b -= c; b -= a; b ^= (a << 10); \
  c -= a; c -= b; c ^= (b >> 15); \
}

int pgc_hash (char *k, int table_size) {
  unsigned int a, b, c, len, length;

  /* Set up the internal state */
  len = strlen (k);
  length = len;
  a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
  c = 0;               /* the previous hash value */
  
  while (len >= 12) {
    a += (k[0] + ((unsigned int) k[1] << 8) + ((unsigned int) k[2] << 16) +
	  ((unsigned int) k[3] << 24));
    b += (k[4] + ((unsigned int) k[5] << 8) + ((unsigned int) k[6] << 16) +
	  ((unsigned int) k[7] << 24));
    c += (k[8] + ((unsigned int) k[9] << 8) + ((unsigned int) k[10] << 16) +
	  ((unsigned int) k[11] << 24));
    PGC_HASH_MIX (a, b, c);
    k += 12; 
    len -= 12;
  }
  
  c += length;
  switch (len) {              /* all the case statements fall through */
  case 11: c += ((unsigned int) k[10] << 24);
  case 10: c += ((unsigned int) k[9] << 16);
  case 9: c += ((unsigned int) k[8] << 8);
  case 8: b += ((unsigned int) k[7] << 24);
  case 7: b += ((unsigned int) k[6] << 16);
  case 6: b += ((unsigned int) k[5] << 8);
  case 5: b += k[4];
  case 4: a += ((unsigned int) k[3] << 24);
  case 3: a += ((unsigned int) k[2] << 16);
  case 2: a += ((unsigned int) k[1] << 8);
  case 1: a += k[0];
  default:
    break;
  }

  PGC_HASH_MIX (a, b, c);

  /* Optimally, here we'd check to see we had to do this... or whether
     PGC_MAX_BINDINGS was too big? */
  return (c & hashmask ((int) (log ((double) table_size))));
}
