/* Include file that defines the parse tree structure needed by the parser (to
   build the parse tree) and the interpreter (to read / type check) */

#ifndef PGC_PARSER_H
#define PGC_PARSER_H

#define yylex yy_exprlex

struct pgc_parse_tree {
  int node_type;
  int node_value_type;
  int line_number;
  void *node_value;
  int num_children;
  struct pgc_parse_tree **children;
  /* Allows us to maintain a pointer to the head of a list... */
  struct pgc_parse_tree *head_pointer;
};

/* Explicit prototype for yylex... */
extern int yylex ();

/* Contains various useful messages for whoever's traversing the tree... */
enum {
  PGC_PARSE_EXPR,
  PGC_PARSE_EXPRESSION_SEQUENCE,
  PGC_PARSE_ARGUMENT_SEQUENCE,

  PGC_PARSE_BINOP,
  PGC_PARSE_BINOP_PLUS,  
  PGC_PARSE_BINOP_MINUS,  
  PGC_PARSE_BINOP_MULT,
  PGC_PARSE_BINOP_DIV,
  PGC_PARSE_BINOP_MOD,
  PGC_PARSE_BINOP_POWER,
  PGC_PARSE_BINOP_EQ,
  PGC_PARSE_BINOP_NEQ,
  PGC_PARSE_BINOP_LT,
  PGC_PARSE_BINOP_GT,
  PGC_PARSE_BINOP_LTE,
  PGC_PARSE_BINOP_GTE,

  PGC_PARSE_VALUE,
  PGC_PARSE_VALUE_INT,
  PGC_PARSE_VALUE_FLOAT,
  PGC_PARSE_VALUE_STRING,
  PGC_PARSE_VALUE_NIL,
  PGC_PARSE_VALUE_VECTOR,
  PGC_PARSE_VALUE_RESULT,

  PGC_PARSE_FUNCTION, /* 23 */

  PGC_PARSE_LVALUE,
  PGC_PARSE_LVALUE_WITH_SUBSCRIPT,
  PGC_PARSE_LVALUE_WITH_FIELD,
  PGC_PARSE_LVALUE_IDENTIFIER,

  PGC_PARSE_ASSIGNMENT,

  PGC_PARSE_IF_STATEMENT,
  PGC_PARSE_IF_ELSE_STATEMENT,

  PGC_PARSE_FOREACH,
  PGC_PARSE_DOWHILE,
  PGC_PARSE_WHILEDO,

  PGC_PARSE_ERROR
};

#endif
