/* Bison file for PGC */
%{
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_parser.h"

  struct pgc_parse_tree *newnode 
    (int node_type, int node_value_type, int line_number,
     void *node_value, int num_children, struct pgc_parse_tree **children) {
    struct pgc_parse_tree *node = 
      (struct pgc_parse_tree *) malloc (sizeof (struct pgc_parse_tree));
    node->node_type = node_type;
    node->node_value_type = node_value_type;
    node->line_number = line_number + 1;
    node->node_value = node_value;
    node->num_children = num_children;
    node->children = children;
    /* For the sake of simplicity... */
    node->head_pointer = NULL;
    return node;
  }

  void yyerror (char *s) {
    fprintf (stderr, "%s\n", s);
    return;
  }

  struct pgc_parse_tree *pgc_root = NULL;
  /* yydebug = 1; */

  /* These declarations will be used by most of the rules -- but we have to
     remember to initialize them every time.  Who knows what the values could
     be otherwise? */
  struct pgc_parse_tree **children = NULL;
  int num_children;
  void *value;

  extern int yy_exprlineno;

#define YYERROR_VERBOSE 1
#define YYDEBUG 1

%}

%union {
  int intval;
  double floatval;
  char *stringval;
  struct pgc_parse_tree *treenode;
}

%start program
%debug

%token <intval> PGC_LEX_INT 
%token <floatval> PGC_LEX_FLOAT
%token <stringval> PGC_LEX_STRING
%token PGC_LEX_NIL

%token PGC_LEX_LPAREN
%token PGC_LEX_RPAREN
%token PGC_LEX_RBRACK

%token <intval> PGC_LEX_ERROR
%token <intval> PGC_LEX_ERROR_EOF

%token PGC_LEX_IF
%token PGC_LEX_ELSE
%token PGC_LEX_FOREACH PGC_LEX_IN
%token PGC_LEX_COMMA
%token PGC_LEX_DOT
%token PGC_LEX_SEMICOLON

%nonassoc PGC_LEX_COMMA
%nonassoc PGC_LEX_SEMICOLON

%nonassoc <stringval> PGC_LEX_IDENTIFIER
%nonassoc PGC_LEX_LBRACK PGC_LEX_LPAREN
%nonassoc PGC_LEX_WHILE PGC_LEX_DO PGC_LEX_ASSIGN
%nonassoc PGC_LEX_BINOP_NEQ PGC_LEX_BINOP_EQ PGC_LEX_BINOP_LT PGC_LEX_BINOP_GT
	PGC_LEX_BINOP_LTE PGC_LEX_BINOP_GTE
%left PGC_LEX_BINOP_PLUS PGC_LEX_BINOP_MINUS
%left PGC_LEX_BINOP_MULT PGC_LEX_BINOP_DIV
%left PGC_LEX_BINOP_MOD
%right PGC_LEX_BINOP_POWER
%nonassoc PGC_LEX_RPAREN

%type <treenode> additionalarguments argumentsequence additionalexprs
	exprsequence lvalue expr program

%%

program: exprsequence {
  pgc_root = $1;
};

/* argument list handling... */
additionalarguments: additionalarguments PGC_LEX_COMMA expr {
  children = malloc (sizeof (struct pgc_parse_tree *));
  $$ = newnode (PGC_PARSE_ARGUMENT_SEQUENCE,
		PGC_PARSE_EXPR, @1.first_line, $3, 0, NULL);
  if ($1 != NULL) {
    children[0] = $$;
    $1->num_children = 1;
    $1->children = children;
    $$->head_pointer = $1->head_pointer;
  }
  else
    $$->head_pointer = $$;
};
               | {$$ = NULL;};

argumentsequence: expr additionalarguments {
  num_children = 0;
  children = NULL;
  if ($2 != NULL) {
    num_children = 1;
    children = malloc (sizeof (struct pgc_parse_tree *));
    children[0] = $2->head_pointer;
  }
  $$ = newnode (PGC_PARSE_ARGUMENT_SEQUENCE, 
		PGC_PARSE_EXPR, @1.first_line, $1, num_children, children);
};
            | {
  $$ = newnode (PGC_PARSE_ARGUMENT_SEQUENCE,
		PGC_PARSE_VALUE_NIL, -1, NULL, 0, NULL);
};

additionalexprs: additionalexprs PGC_LEX_SEMICOLON expr {
  children = malloc (sizeof (struct pgc_parse_tree *));
  $$ = newnode (PGC_PARSE_EXPRESSION_SEQUENCE,
		PGC_PARSE_EXPR, @1.first_line, $3, 0, NULL);
  if ($1 != NULL) {
    children[0] = $$;
    $1->num_children = 1;
    $1->children = children;
    $$->head_pointer = $1->head_pointer;
  }
  else
    $$->head_pointer = $$;
};
               | {$$ = NULL;};

exprsequence: expr additionalexprs {
  num_children = 0;
  children = NULL;
  if ($2 != NULL) {
    num_children = 1;
    children = malloc (sizeof (struct pgc_parse_tree *));
    children[0] = $2->head_pointer;
  }
  $$ = newnode (PGC_PARSE_EXPRESSION_SEQUENCE, 
		PGC_PARSE_EXPR, @1.first_line, $1, num_children, children);
};
            | {
  $$ = newnode (PGC_PARSE_EXPRESSION_SEQUENCE,
		PGC_PARSE_VALUE_NIL, -1, NULL, 0, NULL);
};


/* lvalue handling... */
lvalue: PGC_LEX_IDENTIFIER {
  (char *) value = strdup ($1);
  $$ = newnode (PGC_PARSE_LVALUE, PGC_PARSE_LVALUE_IDENTIFIER, 
		@1.first_line, value, 0, NULL);
  $$->head_pointer = $$;
};
      | lvalue PGC_LEX_DOT PGC_LEX_IDENTIFIER {
	(char *) value = strdup ($3);
	$$ = newnode (PGC_PARSE_LVALUE, PGC_PARSE_LVALUE_WITH_FIELD, 
		      @1.first_line, value, 0, NULL);
	$1->num_children = 1;
	$1->children = malloc (sizeof (struct pgc_parse_tree **));
	$1->children[0] = $$;
	$$->head_pointer = $1->head_pointer;
      };
      | lvalue PGC_LEX_LBRACK expr PGC_LEX_RBRACK {
	/* In this case, "value" is the parse tree for expr... */
	$$ = newnode (PGC_PARSE_LVALUE, PGC_PARSE_LVALUE_WITH_SUBSCRIPT, 
		      @1.first_line, $3, 0, NULL);
	$1->num_children = 1;
	$1->children = malloc (sizeof (struct pgc_parse_tree **));
	$1->children[0] = $$;
	$$->head_pointer = $1->head_pointer;
      };

/* Semantic actions for the binary operators follow. */
expr: expr PGC_LEX_BINOP_PLUS expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_PLUS, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_MINUS expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_MINUS, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_MULT expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_MULT, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_DIV expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_DIV, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_MOD expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_MOD, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_POWER expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_POWER, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_EQ expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_EQ, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_NEQ expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_NEQ, @1.first_line,
		NULL, 2, children);
};
expr: expr PGC_LEX_BINOP_LT expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_LT, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_GT expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_GT, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_LTE expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_LTE, @1.first_line,
		NULL, 2, children);
};

expr: expr PGC_LEX_BINOP_GTE expr {
  children = 
    (struct pgc_parse_tree **) malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $1;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_BINOP, PGC_PARSE_BINOP_GTE, @1.first_line,
		NULL, 2, children);
};

/* Semantic actions for the straight-up values follow. */
expr: PGC_LEX_INT {
  (int *) value = (int *) malloc (sizeof (int));
  *((int *) value) = yylval.intval;
  $$ = newnode 
    (PGC_PARSE_VALUE, PGC_PARSE_VALUE_INT, @1.first_line,
     (void *) value, 0, NULL); 
};

expr: PGC_LEX_FLOAT {
  (double *) value = (double *) malloc (sizeof (double));
  *((double *) value) = yylval.floatval;
  $$ = newnode 
    (PGC_PARSE_VALUE, PGC_PARSE_VALUE_FLOAT, @1.first_line,
     (void *) value, 0, NULL);
};

expr: PGC_LEX_NIL {
  $$ = newnode (PGC_PARSE_VALUE, PGC_PARSE_VALUE_NIL, @1.first_line,
		NULL, 0, NULL); 
};

expr: PGC_LEX_STRING {
  (char *) value = strdup (yylval.stringval);
  $$ = newnode 
    (PGC_PARSE_VALUE, PGC_PARSE_VALUE_STRING, @1.first_line,
     (void *) value, 0, NULL);
};

/* These are vectors... */
expr: PGC_LEX_LBRACK argumentsequence PGC_LEX_RBRACK {
  children = malloc (sizeof (struct pgc_parse_tree *));
  children[0] = $2;
  $$ = newnode (PGC_PARSE_VALUE, PGC_PARSE_VALUE_VECTOR, @1.first_line,
		NULL, 1, children);
};

expr: lvalue {
  /* Otherwise, we'll be reading the lvalue backwards! */
  $$ = $1->head_pointer;
};

expr: lvalue PGC_LEX_ASSIGN expr {
  children = 
    (struct pgc_parse_tree **) (malloc (sizeof (struct pgc_parse_tree *) * 2));
  /* Otherwise, we'll be reading the lvalue backwards! */
  children[0] = $1->head_pointer;
  children[1] = $3;
  $$ = newnode (PGC_PARSE_ASSIGNMENT, 0, @1.first_line, NULL, 2, children);
};

expr: PGC_LEX_LPAREN exprsequence PGC_LEX_RPAREN {
  if ($2 == NULL)
    $$ = newnode (PGC_PARSE_VALUE, PGC_PARSE_VALUE_NIL, @1.first_line, NULL, 0,
		  NULL);
  else 
    $$ = $2;
};

/* Function calls... */
expr: PGC_LEX_IDENTIFIER PGC_LEX_LPAREN argumentsequence PGC_LEX_RPAREN {
  $3->line_number = @2.first_line + 1;
  children = malloc (sizeof (struct pgc_parse_tree **));
  children[0] = $3;
  value = strdup ($1);
  $$ = newnode (PGC_PARSE_FUNCTION, PGC_PARSE_VALUE_NIL, @1.first_line, value, 
		1, children);
};

expr: PGC_LEX_IF PGC_LEX_LPAREN expr PGC_LEX_RPAREN 
      PGC_LEX_LPAREN expr PGC_LEX_RPAREN {
  children = malloc (sizeof (struct pgc_parse_tree **));
  children[0] = $6;
  value = $3;
  $$ = newnode (PGC_PARSE_IF_STATEMENT, PGC_PARSE_EXPR, @1.first_line, value, 
		1, children);
};
expr: PGC_LEX_IF PGC_LEX_LPAREN expr PGC_LEX_RPAREN
      PGC_LEX_LPAREN expr PGC_LEX_RPAREN PGC_LEX_ELSE
      PGC_LEX_LPAREN expr PGC_LEX_RPAREN {
  children = malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $6;
  children[1] = $10;
  value = $3;
  $$ = newnode (PGC_PARSE_IF_ELSE_STATEMENT, PGC_PARSE_EXPR, @1.first_line,
		value, 2, children);
};
expr: PGC_LEX_FOREACH PGC_LEX_IDENTIFIER PGC_LEX_IN PGC_LEX_LPAREN expr
      PGC_LEX_RPAREN PGC_LEX_LPAREN expr PGC_LEX_RPAREN {
  value = strdup ($2);
  children = malloc (sizeof (struct pgc_parse_tree **) * 2);
  children[0] = $5;
  children[1] = $8;
  $$ = newnode (PGC_PARSE_FOREACH, PGC_PARSE_VALUE, @1.first_line, value, 2, 
		children);
};

expr: PGC_LEX_DO PGC_LEX_LPAREN expr PGC_LEX_RPAREN 
      PGC_LEX_WHILE PGC_LEX_LPAREN expr PGC_LEX_RPAREN {
  value = $7;
  children = malloc (sizeof (struct pgc_parse_tree **) * 1);
  children[0] = $3;
  $$ = newnode (PGC_PARSE_DOWHILE, PGC_PARSE_VALUE, @1.first_line, value, 1, 
		children);
};
expr: PGC_LEX_WHILE PGC_LEX_LPAREN expr PGC_LEX_RPAREN
      PGC_LEX_DO PGC_LEX_LPAREN expr PGC_LEX_RPAREN {
  value = $3;
  children = malloc (sizeof (struct pgc_parse_tree **) * 1);
  children[0] = $7;
  $$ = newnode (PGC_PARSE_WHILEDO, PGC_PARSE_VALUE, @1.first_line, value, 1, 
		children);
};

/* Error handling...  These nodes have a node_value_type that comes from the
   pgc.h enum pgc_error_types... */
expr: error {
  $$ = newnode (PGC_PARSE_ERROR, PGC_ERROR_SYNTAX, @1.first_line, NULL, 0, 
		NULL);
};

expr: PGC_LEX_ERROR {
  $$ = newnode (PGC_PARSE_ERROR, PGC_ERROR_LEXICAL, @1.first_line, NULL, 0, 
		NULL);
};

expr: PGC_LEX_ERROR_EOF {
  $$ = newnode (PGC_PARSE_ERROR, PGC_ERROR_UNEXPECTED_EOF, @1.first_line, NULL,
		0, NULL);
};

%%
