#include <config.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "pgc.h"
#include "pgc_hash.h"
#include "pgc_record.h"

struct pgc_record_definition **record_table = NULL;

void pgc_add_record_type (char *record_type_name, int num_fields, ...) {
  va_list names_and_types;
  int i;
  int hashed_index = pgc_hash (record_type_name, PGC_RECORD_TABLE_SIZE);
  int insertion_index = -1;
  for (i = hashed_index; i < PGC_FUNCTION_TABLE_SIZE; i++) {
    if (record_table[i] == NULL) {
      insertion_index = i;
      break;
    }
    /* Record types should not be getting redefined... */
    if (strcmp (record_type_name, record_table[i]->record_type_name) == 0) 
      return;
  }
  if (insertion_index == -1)
    for (i = 0; i < hashed_index; i++) {
      if (record_table[i] == NULL) {
	insertion_index = i;
	break;
      }
      /* As above... */
      if (strcmp (record_type_name, record_table[i]->record_type_name) == 0)
	return;
    }
  if (insertion_index != -1) {
    record_table[insertion_index] = 
      malloc (sizeof (struct pgc_record_definition));
    record_table[insertion_index]->record_type_name = 
      strdup (record_type_name);
    record_table[insertion_index]->field_count = num_fields;
    if (num_fields == 0) {
      record_table[insertion_index]->public_field_names = NULL;
      record_table[insertion_index]->public_field_types = NULL;
    }
    else {
      int j;
      record_table[insertion_index]->public_field_names = 
	malloc (sizeof (char *) * num_fields);
      record_table[insertion_index]->public_field_types =
	malloc (sizeof (int) * num_fields);
      va_start (names_and_types, num_fields);
      for (j = 0; j < num_fields; j++)
	record_table[insertion_index]->public_field_names[j] = 
	  strdup (va_arg (names_and_types, char *));
      for (j = 0; j < num_fields; j++)
	record_table[insertion_index]->public_field_types[j] =
	  va_arg (names_and_types, int);
      va_end (names_and_types);
    }
  }
  return;
}

struct pgc_record_definition *pgc_lookup_record (char *record_type_name) {
  int i;
  int hashed_index = pgc_hash (record_type_name, PGC_RECORD_TABLE_SIZE);
  for (i = hashed_index; i < PGC_RECORD_TABLE_SIZE; i++) {
    if (record_table[i] == NULL)
      return NULL;
    if (strcmp (record_type_name, record_table[i]->record_type_name) == 0)
      return record_table[i];
  }
  for (i = 0; i < hashed_index; i++) {
    if (record_table[i] == NULL)
      return NULL;
    if (strcmp (record_type_name, record_table[i]->record_type_name) == 0)
      return record_table[i];
  }
  return NULL;
}

/* To add new record types, should that ever be necessary, this would be the
   function to edit! */
void pgc_init_record_table () {
  record_table = calloc (PGC_RECORD_TABLE_SIZE,
			 sizeof (struct pgc_record_definition *));
  pgc_add_record_type ("nil", 0);

  #ifdef HAVE_LIBPROTEINGEOMETRY
  pgc_add_record_type ("pdb_record", 1, "filename", PGC_STRING);
  #endif
}
